#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "linkify.h"

enum link_type {
	LT_NONE = 0,
	LT_URL = 1,
	LT_EMAIL = 2,
};

enum link_type scan_linkify(const uint8_t *p, const uint8_t **end) {
	const unsigned char *marker = NULL;
	const uint8_t *end_ = p;
	*end = p;

	// Autogenerated tag variables used by the lexer to track tag values.
	/*!stags:re2c format = 'const uint8_t *@@;\n'; */

/*!re2c

	re2c:define:YYCTYPE  = "uint8_t";
	re2c:define:YYCURSOR = p;
	re2c:define:YYMARKER = marker;
	re2c:define:YYCTXMARKER = marker;
	re2c:yyfill:enable = 0;

	alpha       = [a-zA-Z];
	digit       = [0-9];
	hexdigit    = [0-9a-fA-F];
	unreserved  = alpha | digit | [-._~];
	pct_encoded = "%" hexdigit{2};
	sub_delims  = [!$&'()*+,;=];
	pchar       = unreserved | pct_encoded | sub_delims | [:@];

	scheme = alpha (alpha | digit | [-+.])*;
	userinfo = (unreserved | pct_encoded | sub_delims | ":")*;
	dec_octet
		= digit
		| [\x31-\x39] digit
		| "1" digit{2}
		| "2" [\x30-\x34] digit
		| "25" [\x30-\x35];
	ipv4address = dec_octet "." dec_octet "." dec_octet "." dec_octet;
	h16         = hexdigit{1,4};
	ls32        = h16 ":" h16 | ipv4address;
	ipv6address
		=                            (h16 ":"){6} ls32
		|                       "::" (h16 ":"){5} ls32
		| (               h16)? "::" (h16 ":"){4} ls32
		| ((h16 ":"){0,1} h16)? "::" (h16 ":"){3} ls32
		| ((h16 ":"){0,2} h16)? "::" (h16 ":"){2} ls32
		| ((h16 ":"){0,3} h16)? "::"  h16 ":"     ls32
		| ((h16 ":"){0,4} h16)? "::"              ls32
		| ((h16 ":"){0,5} h16)? "::"              h16
		| ((h16 ":"){0,6} h16)? "::";
	ipvfuture   = "v" hexdigit+ "." (unreserved | sub_delims | ":" )+;
	ip_literal  = "[" ( ipv6address | ipvfuture ) "]";
	reg_name    = (unreserved | pct_encoded | sub_delims)*;
	host
		= ip_literal
		| ipv4address
		| reg_name;
	port      = digit*;
	authority = (userinfo "@")? host (":" port)?;
	path_abempty  = ("/" pchar*)*;
	path_absolute = "/" (pchar+ ("/" pchar*)*)?;
	path_rootless = pchar+ ("/" pchar*)*;
	path_empty    = "";
	hier_part
		= "//" authority path_abempty;
		// Full grammar includes this, but disabled for this purpose
		// | (path_absolute | path_rootless | path_empty);
	query    = (pchar | [/?])*;
	fragment = (pchar | [/?])*;
	uri = scheme ":" hier_part ("?" query)? ("#" fragment)? @end_;

	local_part = (unreserved | pct_encoded | sub_delims | [:/])*;
	email = local_part "@" unreserved{3,}  @end_ ">"?;

	*   { return LT_NONE; }
	uri { 
		// Do not include trailing dots
		if (end_[-1] == '.') --end_;
		*end = end_;
		return LT_URL;
	}
	email {
		if (end_[-1] == '.') --end_;
		*end = end_;
		return LT_EMAIL;
	}
*/
}

void try_linkify(FILE *output, const uint8_t *buf, const uint8_t **end) {
	enum link_type type = scan_linkify(buf, end);
	switch (type) {
	case LT_NONE:
		return;
	case LT_URL:
		fprintf(output, "<a href=\"%.*s\">%.*s</a>", (int)(*end - buf), buf, (int)(*end - buf), buf);
		break;
	case LT_EMAIL:
		fprintf(output, "<a href=\"mailto:%.*s\">%.*s</a>", (int)(*end - buf), buf, (int)(*end - buf), buf);
		break;
	}
}
